/**
*  Copyright Mikael Hgdahl - triyana@users.sourceforge.net
*
*  This source is distributed under the terms of the Q Public License version 1.0,
*  created by Trolltech (www.trolltech.com).
*/

#include "MHChart.h"
#include "MHDouble.h"
#include "MHDC.h"
#include "MHDebug.h"
#include <math.h>



/**
*  Create new chart.
*/
MHChart::MHChart () {
    for (int i = 0; i < MAX_SERIES; i++)
        aPoints[i] = 0;
    Clear ();
}



/**
*  Destroy object
*/
MHChart::~MHChart () {
    Clear ();
}



/**
*  Add new list of chart data to draw. This method is called by derived class.
*  @param const char* - Name
*  @param MHVector*   - Vector with price data in the form of MHPoint
*  @param bool        - default false which says that the y chart labels is going to the right
*  @return int        - MHChart::RIGHT or MHChart::LEFT
*  @see MHPoint
*/
bool MHChart::AddData (const char* name, int type, MHVector* ve, int align) {
    aOrgSize = 0;
    if (aSerieCount == MAX_SERIES)
        return false;

    aPoints[aSerieCount]    = ve;
    aLabels[aSerieCount]    = name;
    aTypes[aSerieCount]     = type;
    aPositions[aSerieCount] = align;

    //  Save max range
    if (ve->Size() > aRangeMax)
        aRangeMax = ve->Size();

    aOrgSize = aRangeMax;
    aSerieCount++;
    return true;
}



/**
*  Remove all data sets and reset settings.
*/
void MHChart::Clear() {
    aSettings.Clear ();
    aInit               = false;
    aXDefStep           = 0;
    aRangeMax           = 0;
    aRangeStart         = 0;
    aRangeSize          = 0;
    aCurrentY           = RIGHT;
    aYMin[LEFT]         = 0;
    aYMin[RIGHT]        = 0;
    aYMax[LEFT]         = 0;
    aYMax[RIGHT]        = 0;
    aYStep[LEFT]        = 0;
    aYStep[RIGHT]       = 0;
    aYPixelRatio[LEFT]  = 0;
    aYPixelRatio[RIGHT] = 0;
    aValid[LEFT]        = 0;
    aValid[RIGHT]       = 0;
    aXMin               = 0;
    aXMax               = 0;
    aXSpan              = 0;
    aXStep              = 0;
    aXPixelRatio        = 0;

    for (int i = 0; i < MAX_SERIES; i++) {
        aPositions[i] = -1;
        if (aPoints[i]) {
            aPoints[i]->Erase();
            delete aPoints[i];
            aPoints[i] = 0;
        }
    }
    aSerieCount = 0;
}



/**
*  Print var- values to debug file
*/
void MHChart::Debug () {
    DBG ("MHChart::Debug")

    DBG_LOG1 ("aInit               = %d", aInit)
    DBG_LOG1 ("aRangeSize          = %d", aRangeSize)
    DBG_LOG1 ("aRangeMax           = %d", aRangeMax)
    DBG_LOG1 ("aRangeStart         = %d", aRangeStart)
    DBG_LOG1 ("aCanvasWidth        = %f", aCanvasWidth)
    DBG_LOG1 ("aCanvasHeight       = %f", aCanvasHeight)
    DBG_LOG1 ("aYMin[LEFT]         = %f", aYMin[LEFT])
    DBG_LOG1 ("aYMin[RIGHT]        = %f", aYMin[RIGHT])
    DBG_LOG1 ("aYMax[LEFT]         = %f", aYMax[LEFT])
    DBG_LOG1 ("aYMax[RIGHT]        = %f", aYMax[RIGHT])
    DBG_LOG1 ("aYStep[LEFT]        = %f", aYStep[LEFT])
    DBG_LOG1 ("aYStep[RIGHT]       = %f", aYStep[RIGHT])
    DBG_LOG1 ("aYPixelRatio[LEFT]  = %f", aYPixelRatio[LEFT])
    DBG_LOG1 ("aYPixelRatio[RIGHT] = %f", aYPixelRatio[RIGHT])
    DBG_LOG1 ("aXMin               = %f", aXMin)
    DBG_LOG1 ("aXMax               = %f", aXMax)
    DBG_LOG1 ("aXSpan              = %f", aXSpan)
    DBG_LOG1 ("aXStep              = %f", aXStep)
    DBG_LOG1 ("aXPixelRatio        = %f", aXPixelRatio)
}



/**
*  Copyrigth (c) 2000, Michael P.D Bramley.
*  Calc axis number.
*  @param double - min value
*  @param double - max value
*  @param double - step value
*/
void MHChart::defineAxis (int align) {
    DBG ("MHChart::DefineAxis")
    DBG_LOG1 ("aYMin[align]  =%f", aYMin[align])
    DBG_LOG1 ("aYMax[align]  =%f", aYMax[align])
    DBG_LOG1 ("aYStep[align] =%f", aYStep[align])

    double Test_inc, Test_min, Test_max, Range;
    int    i = 0;

    Range = aYMax[align] - aYMin[align];
    if (Range < 0) {
        aYStep[align] = 0;
        return;
    }
    else if (Range < 0.0000001) {
        aYMin[align] = ceil (aYMax[align]) - 1;
        aYMax[align] = aYMin[align] + 1;
        aYStep[align] = 1;
        return;
    }

    Test_inc = pow (10, ceil (log10 (Range / 10)));
    Test_max = ((long)(aYMax[align] / Test_inc)) * Test_inc;

    if (Test_max < aYMax[align])
        Test_max += Test_inc;

    Test_min = Test_max;

    do {
        ++i;
        Test_min -= Test_inc;
    } while (Test_min > aYMin[align]);

    if (i < 6) {
        Test_inc /= 2;
        if ((Test_min + Test_inc) <= aYMin[align])
            Test_min += Test_inc;
        if ((Test_max - Test_inc) >= aYMax[align])
            Test_max -= Test_inc;
    }

    aYMin[align]        = Test_min;
    aYMax[align]        = Test_max;
    aYStep[align]       = Test_inc;
    aYPixelRatio[align] = aCanvasHeight / (aYMax[align] - aYMin[align]);

    DBG_LOG1 ("aYMin[align]         =%f", aYMin[align])
    DBG_LOG1 ("aYMax[align]         =%f", aYMax[align])
    DBG_LOG1 ("aYStep[align]        =%f", aYStep[align])
    DBG_LOG1 ("aYPixelRatio[align]  =%f", aYPixelRatio[align])
}



/**
*  Draw labels for y range.
*  @param int - Label index
*  @param int - Color from enum MHChart::COLORS
*/
void MHChart::DrawLabel (int nIndex, int nColor, int nRow, int labelLen) {
    aDC->SetBrushColor (nColor);
    aDC->SetPenColor (nColor);
    if (aPositions[nIndex] == LEFT)
        aDC->DrawText (aLabels[nIndex](), aX + aXOffset + aCharWidth, aY + (aCharHeight * (nRow + 1)));
    else
        aDC->DrawText (aLabels[nIndex](), aX + aXOffset - aCharWidth + int(aCanvasWidth) - (labelLen * aCharWidth), aY + (aCharHeight * (nRow + 1)));
}



/**
*  Draw line between two world coordinates.
*  @param double   - X1 Startpoint
*  @param double   - Y1 Startpoint
*  @param double   - X2 Endpoint
*  @param double   - Y2 Endpoint
*  @return double  - X2 Endpoint
*/
double MHChart::DrawLine(double X1, double Y1, double X2, double Y2) {
    X1 = TranslateX(X1);
    Y1 = TranslateY(Y1);
    X2 = TranslateX(X2);
    Y2 = TranslateY(Y2);

    aDC->DrawLine((int)(X1),(int)(Y1),(int)(X2),(int)(Y2));
    return X2;
}



/**
*  Draw rectangle between two pairs of world coordinates.
*  @param double   - X1 Startpoint
*  @param double   - Y1 Startpoint
*  @param double   - X2 Endpoint
*  @param double   - Y2 Endpoint
*  @param bool     - true if rectangle is going to be filled with current color
*/
double MHChart::DrawRect (double X1, double Y1, double X2, double Y2, bool flag) {
    X1 = TranslateX (X1);
    Y1 = TranslateY (Y1);
    X2 = TranslateX (X2);
    Y2 = TranslateY (Y2);

    if (flag == true)
       aDC->DrawRect (int(X1 - 1.0), int(Y2), int(X2 - X1 - 1.0), int(Y1 - Y2));
    return X2;
}



/**
*  Draw text to canvas.
*  @param const char*   - Text so be drawed
*  @param double        - X position
*  @param double        - Y position
*/
void MHChart::DrawText (const char* s, double X, double Y) {
    X = TranslateX (X);
    Y = TranslateY (Y);

    aDC->DrawText (s, (int)(X), (int)(Y));
}



/**
*  Draw y labels for the chart.
*  @param bool - true to draw horisontal lines
*/
void MHChart::DrawYLabels (bool bDrawBox) {
    double       nCurrY        = aYMin[aCurrentY];
    double       nOldY         = -1;
    double       nXTick        = SMALL_TICK_LINE_SIZE / aXPixelRatio;
    double       nAdjustHeight = double(aCharHeight) / (2 * aYPixelRatio[aCurrentY]);
    double       alignText     = aYLabelWidth / aCharWidth;
    double       ydiff         = aCharHeight;
    double       tmp;
    MHString     s;
    MHDouble     d;

    aDC->SetPenColor (BLACK);
    aDC->SetBrushColor (BLACK);

    if (aYStep[aCurrentY] < 1000)
        d.SetSize (aYMax[aCurrentY], aYStep[aCurrentY]);
    else if (aYStep[aCurrentY] < 1000000)
        d.SetSize (aYMax[aCurrentY] / 1000, aYStep[aCurrentY] / 1000);
    else
        d.SetSize (aYMax[aCurrentY] / 1000000, aYStep[aCurrentY] / 1000000);

    if (bDrawBox) {
        DrawLine (aXMin, aYMin[aCurrentY], aXMax, aYMin[aCurrentY]);
        DrawLine (aXMin, aYMax[aCurrentY], aXMax, aYMax[aCurrentY]);
    }

    if (aYMin[aCurrentY] > 2000000000) return;

    while (nCurrY <= aYMax[aCurrentY]) {
        if (aSettings.aYLabels) {
            //  Check if y values are to big to be painted, if so reduce them
            if (aYStep[aCurrentY] < 1000)
                d.ToString(nCurrY, &s);
            else if (aYStep[aCurrentY] < 1000000) {
                tmp = nCurrY;
                tmp /= 1000;
                d.ToString(tmp, &s);
                s += "'";
            }
            else {
                tmp = nCurrY;
                tmp /= 1000000;
                d.ToString(tmp, &s);
                s += "\"";
            }
            s.AlignRight (int(alignText), ' ');

            //  Draw text only if enough y distanse exist between the text
            tmp = TranslateY (nCurrY - nAdjustHeight);
            if (nOldY < 0 || tmp < (nOldY - ydiff)) {
                if (aCurrentY == RIGHT)
                    aDC->DrawText (s(), (int)(aX + aCanvasWidth + aXOffset + aCharWidth + aCharWidth), (int)(TranslateY(nCurrY - nAdjustHeight)));
                else
                    aDC->DrawText (s(), (int)(aX + aCharWidth), (int)(TranslateY(nCurrY - nAdjustHeight)));
                nOldY = tmp;
            }
        }

        //  Draw surrounding box and small tick lines
        if (bDrawBox && aCurrentY == RIGHT && aSettings.aHLines)
            DrawLine (aXMin, nCurrY, aXMax, nCurrY);

        //  Draw ticks
        if (aCurrentY == RIGHT)
            DrawLine (aXMax, nCurrY, aXMax + nXTick, nCurrY);
        else
            DrawLine (aXMin - nXTick, nCurrY, aXMin, nCurrY);

        nCurrY += aYStep[aCurrentY];
    }
}



/**
*  Paint all dataseries and labels
*  @param MHDC* - Device context
*/
void MHChart::Paint (MHDC* pDC) {
    DBG_TIME ("MHChart::Paint")

    bool Init = false;
    bool yInit[2];
    bool box = true;
    int  nRow[2];
    int  i;
    int  labelLen = 0;
    
    Debug ();

    aDC = pDC;
    yInit[0] = yInit[1] = false;
    nRow[0] = nRow[1] = 0;
    
    //  Recalc must initiate before drawing
    if (aInit == false)
        return;

    //  Find out maximum width of right chart labels
    for (i = 0; i < MAX_SERIES; i++) {
        if (aPoints[i]) {
            if (aPositions[i] == RIGHT && aLabels[i].Size() > labelLen)
                labelLen = aLabels[i].Size();
        }
    }
    
    //  Set color of first chart
    int color = BLUE;
    for (i = 0; i < MAX_SERIES; i++) {
        aCurrentY = aPositions[i];

        if (aPoints[i]) {
            if (aValid[aCurrentY] == 1) {
                if (aPositions[i] == LEFT)
                    nRow[LEFT]++;
                else
                    nRow[RIGHT]++;
                
                DBG_LOG3 ("serie=%d  size=%d  color=%d", i, aPoints[i]->Size(), color)
                //  First chart draws x labels
                if (Init == false) {
                    DrawXLabels ();
                    Init = true;
                }
    
                if (yInit[aCurrentY] == false) {
                    DrawYLabels (box);
                    box = false;
                    yInit[aCurrentY] = true;
                }
    
                aDC->BeginClip (int(TranslateX(aXMin)), int(TranslateY(aYMax[aCurrentY])), int(TranslateX(aXMax)), int(TranslateY(aYMin[aCurrentY])));
                DrawChart (i, color);
                DrawLabel (i, color, nRow[aPositions[i]], labelLen);
                aDC->EndClip ();
                color++;
            }
        }
        else
            break;
    }
}



/**
*  Get y value with help of screen coordinate.
*  @param int      - X pos
*  @param int      - Y Pos
*  @param MHPoint& - Left result point
*  @param MHPoint& - Right result point
*  @return bool    - true if set
*/
bool MHChart::QueryLeft (int x, int y, MHPoint* point) {
    if (aSettings.aChartLeft) {
        aCurrentY   = LEFT;
        double x2   = TranslateX2World(x);
        double y2   = TranslateY2World(y);
        double XMin = TranslateX(aXMin);
        double XMax = TranslateX(aXMax);
        double YMin = TranslateY(aYMin[aCurrentY]);
        double YMax = TranslateY(aYMax[aCurrentY]);

        if (y >= YMax && y <= YMin && x >= XMin && x <= XMax) {
            point->SetX (x2);
            point->SetHigh (YMax);
            point->SetLow (YMin);
            point->SetClose (y2);
            return true;
        }
    }
    return false;
}



/**
*  Get y value with help of screen coordinate.
*  @param int      - X pos
*  @param int      - Y Pos
*  @param MHPoint& - Left result point
*  @param MHPoint& - Right result point
*  @return bool    - true if set
*/
bool MHChart::QueryRight (int x, int y, MHPoint* point) {
    if (aSettings.aChartRight) {
        aCurrentY   = RIGHT;
        double x2   = TranslateX2World(x);
        double y2   = TranslateY2World(y);
        double XMin = TranslateX(aXMin);
        double XMax = TranslateX(aXMax);
        double YMin = TranslateY(aYMin[aCurrentY]);
        double YMax = TranslateY(aYMax[aCurrentY]);

        if (y >= YMax && y <= YMin && x >= XMin && x <= XMax) {
            point->SetX (x2);
            point->SetHigh (YMax);
            point->SetLow (YMin);
            point->SetClose (y2);
            return true;
        }
    }
    return false;
}



/**
*  This method must be called when screen size is changed or
*  another chart type is selected or a dataserie is changed.
*/
void MHChart::Recalc () {
    DBG ("MHChart::Recalc")

    //  Set all magic numbers that are used to calc size and distances
    aInit = false;
    if (aPoints[0]) {
        testRange ();
        GetMinMax ();

        if (aSettings.aChartLeft) defineAxis (LEFT);
        if (aSettings.aChartRight) defineAxis (RIGHT);

        aXSpan = aXMax - aXMin;
        aXPixelRatio = aCanvasWidth / aXSpan;
        SetXTick ();

        aInit = true;
        if (fabs (aXSpan) < (1 / MAX_VALUE)) aInit = false;
        if (fabs (aXStep) < (1 / MAX_VALUE) || fabs (aYStep[LEFT]) < (1 / MAX_VALUE)) aInit = false;
        if (aCanvasHeight < 50 || aCanvasWidth < 50) aInit = false;
        if (aRangeSize == 0) aInit = false;
    }
    else {
        aRangeSize = 0;
    }
}




/**
*  This method must be called when screen size is changed or
*  another chart type is selected or a dataserie is changed
*  @param int   - x pos
*  @param int   - y pos
*  @param int   - width
*  @param int   - height
*  @param MHDC* - Device context
*/
void MHChart::Resize (int x, int y, int w, int h, MHDC *pDC) {
    DBG ("MHChart::Resize")

    double tmp, tmp2;

    aX = x;
    aY = y;
    aW = w;
    aH = h;
    aXOffset = 0;

    if (pDC) {
        aDC = pDC;
        //  Set font height and width in pixels
        aCharWidth  = aDC->GetStringWidth ("X");
        aCharHeight = aDC->GetCharHeight ();

        if (aSettings.aYLabels)
            aYLabelWidth = aCharWidth * YBORDER;
        else
            aYLabelWidth = aCharWidth * (YBORDER / 2);

        if (aSettings.aXLabels) {
            aBorder = (XBORDER * aCharHeight);
            aCanvasHeight = aH - (aBorder + (YOFFSET * aCharHeight));
        }
        else {
            aBorder = aCharHeight;
            aCanvasHeight = aH - (aBorder + (YOFFSET * aCharHeight));
        }

        //  Set width of canvas depending on left/right axis
        tmp2 = XOFFSET * aCharWidth;
        tmp  = aYLabelWidth + tmp2;

        if (aSettings.aChartRight && !aSettings.aChartLeft) {
            //  Only right y labels
            aXOffset = int(tmp2);
            aCanvasWidth = aW - tmp - tmp2;
        }
        else if (!aSettings.aChartRight && aSettings.aChartLeft) {
            //  Only left y labels
            aXOffset = int(tmp);
            aCanvasWidth = aW - tmp - tmp2;
        }
        else {
            //  right and left y labels
            aXOffset = int(tmp);
            aCanvasWidth = aW - tmp - tmp;
        }
        aRangeSize = (int) (aCanvasWidth / aSettings.aXWidth);
    }
}



/**
*  calc.  magic number to use for step between the x-lines.
*/
void MHChart::SetXTick() {
    DBG("MHChart::SetXtick")

    int     XLineCount;
    double  XMax;
    double  XSpan = aXSpan;
    double  nRes = 0;
    double  nFact = 0;

    if (aCanvasWidth > MIN_CANVAS_WIDTH && XSpan > (1 / MAX_VALUE) && XSpan < MAX_VALUE) {
        //  Set max x lines to the number of items if possible
        XLineCount = aRangeSize;

        //  Calc x step between every major xstep
        while (1) {
            nRes = fabs (XSpan / XLineCount);
            nFact = 1;
            DBG_LOG2("XSpan=%.2f | nRes=%.2f", XSpan, nRes)

            if (nRes < (1 / MAX_VALUE))
                return;

            while (nRes > 5) {
                nRes /= 10;
                nFact *= 10;
            }

            while (nRes <= 0.5) {
                nRes *= 10;
                nFact /= 10;
            }

            if (nRes > 4.98)
                nRes = 10.0;
            else if (nRes > 0.98)
                nRes = 5.0;
            else
                nRes = 1;

            nRes *= nFact;

            //    Set new xmax according to how many xticks and the steps between them
            XMax = aXMin + (XLineCount * nRes);

            DBG_LOG3("XMax=%.2f | aXmax=%.2f | XLineCount=%.2f", XMax, aXMax, XLineCount)

            if (XMax < aXMax)
                XSpan *= 1.05;
            else
                break;
        }

        MHDouble d;

        d.SetSize (nRes, nRes);
        XMax = d.Round (aXMax);
        aXStep = nRes;
        if (aXDefStep > aXStep)
            aXStep = aXDefStep;
        aXStart = XMax;

        //  Move start of first xtick so xvalue will be a nice round one based on aXStep
        while ((aXStart - aXStep) >= aXMin)
            aXStart -= aXStep;

        //  and move up
        while (aXStart < aXMin)
            aXStart += aXStep;
    }
    else
        aXStep = 0;
    DBG_LOG2("XStart=%.2f | XStep=%.2f", aXStart, aXStep)
}



/**
*  Test if data range are valid.
*/
void MHChart::testRange () {
    if (aSettings.aShowAll) {
        //  All data points that are going to be drawn
        aRangeStart = 0;
        aRangeSize = aRangeMax;
    }
    else {
        if ((aRangeStart + aRangeSize) > aRangeMax ||
            aRangeStart > aRangeMax)
            aRangeStart = aRangeMax - aRangeSize + 1;

        if (aRangeStart < 1 || aRangeStart > MAX_DATA)
            aRangeStart = 0;

        if (aRangeSize > aRangeMax)
            aRangeSize = aRangeMax;
    }
}



/**
*  Translate X world coordinate to screen coordinate
*  @param  double   - World coordinate to translate
*  @return double   - Screen coordinate
*/
double MHChart::TranslateX (double nX) {
    //  Translate x value into screen coordinates
    nX -= aXMin;
    nX *= aXPixelRatio;
    nX += aX + aXOffset;
    return nX;
}



/**
*  Translate X screen coordinate to world coordinate
*  @param const    - Screen coordinate to translate
*  @return double  - World coordinate
*/
double MHChart::TranslateX2World (double nX) {
    nX -= aX;
    nX -= aXOffset;
    nX /= aXPixelRatio;
    nX += aXMin;
    return nX;
}



/**
*  Translate Y world coordinate to screen coordinate.
*  It also swaps origin of y to bottom at canvas
*  @param  double  - World coordinate to translate
*  @return double  - Screen coordinate
*/
double MHChart::TranslateY (double nY) {
    //  Translate y value into screen coordinates
    nY -= aYMin[aCurrentY];
    nY *= aYPixelRatio[aCurrentY];
    nY += aBorder;

    //  Invert y coordinate so y starts from the bottom at the screen
    nY = (aH + aY) - nY;
    return nY;
}



/**
*  Translate Y screen coordinate to world coordinate
*  @param double    - Screen coordinate to translate
*  @return double   - World coordinate
*/
double MHChart::TranslateY2World (double nY) {
    nY = (aH + aY) - nY;
    nY -= aBorder;
    nY /= aYPixelRatio[aCurrentY];
    nY += aYMin[aCurrentY];

    return nY;
}
