/**
*  Copyright Mikael Hgdahl - triyana@users.sourceforge.net
*
*  This source is distributed under the terms of the Q Public License version 1.0,
*  created by Trolltech (www.trolltech.com).
*/

#include "MHDC.h"
#include "MHChart.h"
#include "MHDate.h"
#include "MHDateChart.h"
#include "MHDebug.h"
#include "MHPoint.h"



/**
*  Create device contect
*/
MHDC::MHDC () :
MH () {
    aW          = 0;
    aH          = 0;
    aFontSize   = 12;
    aChartAreas = new MHVector ();
}



/**
*  Destroy object
*/
MHDC::~MHDC () {
    Clear();
    delete aChartAreas;
    aDates.Erase ();
}



/**
*  Add a new MHChart object this canvas
*  @param MHChart* - The chart object
*  @param double   - X pos, the number is between 0-1
*  @param double   - Y pos, the number is between 0-1
*  @param double   - Width, the number is between 0-1
*  @param double   - Height, the number is between 0-1
*/
void MHDC::Add (MHChart* pChart, double x, double y, double w, double h) {
    MHChartArea* m = new MHChartArea ();

    m->aChart = pChart;
    m->aW     = w;
    m->aH     = h;
    m->aX     = x;
    m->aY     = y;

    aChartAreas->Push (m);
}



/**
*  Align all chart areas so they have the same size.
*  @param int - Alignment
*/
void MHDC::Align (int align) {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea*     c = (MHChartArea*) aChartAreas->At (f);
        MHChartSettings* s = c->aChart->GetSettings();

        if (align == MHChart::LEFT)
            s->aChartLeft = 1;
        else if (align == MHChart::RIGHT)
            s->aChartRight = 1;
    }
}



/**
*  Destroy all charts
*/
void MHDC::Clear () {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

        delete c->aChart;
        c->aChart = 0;
    }
    aChartAreas->Erase();
}



/**
*  Create the timeserie that rules the x pos for the date charts.
*  Call it only for the MHDateChart object
*  @param int       - The unique id that the MHChart object can identify aelf width
*  @return MHChart* - An instance of MHChart or 0 if object can't be found
*/
void MHDC::CreateTimeSerie() {
    DBG ("MHDC::CreateTimeSerie")

    int        f;
    double     size = 0;
    MHString   min = "20300101";
    MHString   max = "19000101";
    MHString   tmp, tmp2;

    aDates.Erase ();
    if (aChartAreas->Size() == 0) return;
    DBG_LOG1("Class=%s", ((MHChartArea*)aChartAreas->At(0))->aChart->Class())
    if (strcmp(((MHChartArea*)aChartAreas->At(0))->aChart->Class(), "MHDateChart") != 0) return;

    //  Get min, nax dates
    for (f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* a = (MHChartArea*) aChartAreas->At (f);
        MHDateChart* c = (MHDateChart*) a->aChart;

        c->PopDates ();
        c->GetDateRange (&tmp, &tmp2);
        DBG_LOG3 ("start=%s, stop=%s, size=%d", tmp(), tmp2(), c->GetSize ())
        if (tmp < min) min = tmp;
        if (tmp2 > max) max = tmp2;
        if (c->GetSize () > size) size = c->GetSize ();
    }

    DBG_LOG2 ("min=%s, max=%s", min(), max());
    if (size > 0) {
        MHChartArea* a = (MHChartArea*) aChartAreas->At (0);
        MHDateChart* c = (MHDateChart*) a->aChart;

        MHDate d1 (c->GetMinDate()->Get());
        double w = (double(c->GetCanvasWidth()) / size);
        DBG_LOG3 ("canvasWidth=%d, size=%f, tickw=%f", c->GetCanvasWidth (), size, w)
        if (w > c->GetSettings()->aXWidth) {
            DBG_LOG("Forcing range")
            int tmp = int((c->GetCanvasWidth () / c->GetSettings()->aXWidth) - size);
            if (c->GetSettings()->aTimeType == MHDate::FRIDAY)
                d1.AddDays (-(tmp) * 7);
            else
                d1.AddDays (-(tmp));
            d1.Get (MHDate::LONG_DATE, min);
        }

        MHDate::CreateTimeSerie (min(), max(), c->GetSettings()->aTimeType, c->GetClosed(), &aDates);
    }
    else
        return;

    DBG_LOG2 ("min=%s, max=%s", min(), max());
    //  Set new dates and calc date serie
    for (f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* a = (MHChartArea*) aChartAreas->At (f);
        MHDateChart* c = (MHDateChart*) a->aChart;

        c->SetDateRange (min(), max());
        c->SetTimeSerie (&aDates);
    }
}



/**
*  Dump all chart values
*/
void MHDC::Debug () {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

        c->aChart->Debug ();
    }
}


/**
*  Get a MHChart object with the help of id.
*  @param int       - Chart index
*  @return MHChart* - An instance of MHChart or 0 if object can't be found
*/
MHChart* MHDC::GetChart (int id) {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

        if ((f + 1)== id)
            return c->aChart;
    }
    return 0;
}



/**
*  Find out how big is the largest dataset of all MHChart objects
*  @return int - The largest size
*/
int MHDC::GetLargestDataSet () {
    int max   = 0;
    int range = 0;

    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

        if (c->aChart->GetSize() > max)
            max = c->aChart->GetSize();
        if (c->aChart->GetRange() > range)
            range = c->aChart->GetRange();
    }
    return max - range + 1;
}



/**
*  Get settings for a chart.
*  @param int                - Chart index
*  @return MHChartSettings*  - Settings object
*/
MHChartSettings* MHDC::GetSettings (int id) {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);
        if ((f + 1) == id)
            return c->aChart->GetSettings();
    }
    return 0;
}



/**
*  Repaint all chart objects
*/
void MHDC::Paint () {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);
        c->aChart->Paint(this);
    }
}



/**
*  Query data point
*  @param int      - X pos
*  @param int      - Y Pos
*  @param MHPoint& - Result point
*  @param MHDC* - Device context
*  @return bool    - true if found
*/
bool MHDC::QueryLeft (int x, int y, MHPoint* p) {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

        if (c->aChart->QueryLeft (x, y, p) == true)
            return true;
    }
    return false;
}



/**
*  Query data point
*  @param int      - X pos
*  @param int      - Y Pos
*  @param MHPoint& - Result point
*  @param MHDC* - Device context
*  @return bool    - true if found
*/
bool MHDC::QueryRight (int x, int y, MHPoint* p) {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

        if (c->aChart->QueryRight (x, y, p) == true)
            return true;
    }
    return false;
}



/**
*  Recalc chart data
*  @param bool - set to true if data serie has been added/changed
*  @param int  - Width of canvas, default -1, no resize is done
*  @param int  - Height of canvas, default -1, no resize is done
*/
void MHDC::Recalc (bool bInit, int w, int h) {
    DBG ("MHDC::Recalc")
    DBG_LOG2 ("width=%d, height=%d", w, h)

    int f;

    if (bInit) {
        int left  = 0;
        int right = 0;

        for (f = 0; f < aChartAreas->Size(); f++) {
            MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

            for (int f = 0; f < MHChart::MAX_SERIES; f++) {
                if (c->aChart->aPoints[f] == 0)
                    continue;

                if (c->aChart->aPositions[f] == MHChart::LEFT)
                    left = 1;
                else if (c->aChart->aPositions[f] == MHChart::RIGHT)
                    right = 1;
            }
        }
        for (f = 0; f < aChartAreas->Size(); f++) {
            MHChartArea*     c   = (MHChartArea*) aChartAreas->At (f);
            MHChartSettings* set = c->aChart->GetSettings ();
            set->aChartLeft      = left;
            set->aChartRight     = right;
        }
    }

    if (w >= 0 && h >= 0) {
        for (f = 0; f < aChartAreas->Size(); f++) {
            MHChartArea* c = (MHChartArea*) aChartAreas->At (f);

            c->aChart->Resize (
                (int) (w * c->aX),
                (int) (h * c->aY),
                (int) ((w * c->aW) - (w * c->aX)),
                (int) ((h * c->aH) - (h * c->aY)),
                this);
        }
    }

    if (bInit) {
        CreateTimeSerie();
        SetScrollRange ();
    }

    for (f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);
        c->aChart->Recalc ();
    }
}



/**
*  Set start pos in all data sets.
*  @param int - Index in vector for the first point to draw
*/
void MHDC::SetStartInDataSet (int start) {
    for (int f = 0; f < aChartAreas->Size(); f++) {
        MHChartArea* c = (MHChartArea*) aChartAreas->At (f);
        c->aChart->SetStart (start);
    }
}
