/**  Copyright Mikael Hgdahl - triyana@users.sourceforge.net
*
*  This source is distributed under the terms of the Q Public License version 1.0,
*  created by Trolltech (www.trolltech.com).
*/

#include <MHDateChart.h>
#include <MHDate.h>
#include <MHDC.h>
#include <MHDebug.h>
#include <MHPoint.h>



/**
*  Send all params to MHChart's constructor
*  @param int - A unique id number if parent is MHChartManager
*/
MHDateChart::MHDateChart () : MHChart () {
    aClosed  = 0;
    aDates   = 0;
    Clear ();
}



/**
*  Remove all from memory
*/
MHDateChart::~MHDateChart () {
    Clear ();
}



/**
*  Add new list of data to draw on screen
*  @param const char*   - Label
*  @param int           - Type of chart, BAR_CHART, LINE_CHART or VOLUME_CHART
*  @param MHVector*     - Vector with price data in the form of MHPoint
*  @param bool          - default false for adding charts to the right
*  @return int          - MHChart::RIGHT or MHChart::LEFT
*  @see MHPoint
*/
bool MHDateChart::AddData (const char* name, int type, MHVector* ve, int align) {
    DBG ("MHDateChart::AddData")

    MHPoint*   pStart = (MHPoint*) ve->First ();
    MHPoint*   pStop  = (MHPoint*) ve->Last ();

    if (pStart || pStop) {
        DBG_LOG2 ("start=%s, stop=%s", pStart->GetDate(), pStop->GetDate());
        //  Find out min/max date in timeserie
        if (aXMinDate > pStart->GetDate())
            aXMinDate = pStart->GetDate();
        if (aXMaxDate < pStop->GetDate())
            aXMaxDate = pStop->GetDate();
    }
    aXOrgMinDate = aXMinDate;
    aXOrgMaxDate = aXMaxDate;

    DBG_LOG2 ("min=%s, max=%s", aXMinDate(), aXMaxDate());
    return MHChart::AddData (name, type, ve, align);
}



/**
*  Remove all data sets and set some variables to default
*/
void MHDateChart::Clear() {
    if (aClosed) aClosed->Erase ();
    delete aClosed;
    aClosed = 0;
    MHChart::Clear ();
    aXMinDate = "20300101";
    aXMaxDate = "19700101";
}



/**
*  Print values to debug file
*/
void MHDateChart::Debug () {
    DBG ("Debug")

    MHChart::Debug ();

    DBG_LOG1 ("aXMinDate           = %s", aXMinDate())
    DBG_LOG1 ("aXMaxDate           = %s", aXMaxDate())
    #ifdef BUGG
        if (aDates)
            DBG_LOG1 ("aDates          = %d", aDates->Size());
    #endif
}



/**
*  Draw chart for a time serie.
*  @param int - Data serie to draw
*  @param int - One of MHChart::COLORS
*/
void MHDateChart::DrawChart (int nIndex, int nColor) {
    DBG_TIME ("DrawChart")

    int             nCurrentIndex = aRangeStart;
    int             nLastIndex    = aRangeStart + aRangeSize;
    bool            bInit         = false;
    int             foundIndex    = 0;
    int             nPointIndex   = 0;
    double          nX2           = 0;
    double          nY2           = 0;
    MHPoint*        point;
    MHPoint         point2;

    aDC->SetPenColor (nColor);

    //  Iterate through the dateserie
    while (nCurrentIndex < nLastIndex && nCurrentIndex < aDates->Size()) {
        //  Use date from DateSerie and find point
        point2.SetDate ((MHString*)(*aDates)[nCurrentIndex]);

        if (nPointIndex == 0)
            point = (MHPoint*) aPoints[nIndex]->Find (&point2, nPointIndex, 0);
        else {
            point = (MHPoint*) aPoints[nIndex]->Search (&point2, nPointIndex, 0, &foundIndex);
            if (point) nPointIndex = foundIndex;
        }

        if (point) {
            switch (aTypes[nIndex]) {
                case LINE_CHART:
                    if (bInit == true)
                        DrawLine (nCurrentIndex, point->GetClose(), nX2, nY2);
                    nX2 = nCurrentIndex;
                    nY2 = point->GetClose();
                    bInit = true;
                    break;

                case BAR_CHART:
                    DrawLine (nCurrentIndex, point->GetLow(), nCurrentIndex, point->GetHigh());
                    DrawLine (nCurrentIndex, point->GetClose(), nCurrentIndex + 1, point->GetClose());
                    break;

                case RECT_CHART:
                    DrawRect (nCurrentIndex, aYMin[aCurrentY], nCurrentIndex + 1, point->GetClose(), 1);
                    break;

                case VERT_CHART:
                    DrawLine (nCurrentIndex, aYMin[aCurrentY], nCurrentIndex, aYMax[aCurrentY]);
                    break;

                default:
                    break;
            }
        }
        nCurrentIndex++;
    }
}



/**
*  Draw labels for price chart. The labels consists of date strings.
*/
void MHDateChart::DrawXLabels () {
    DBG_TIME ("MHDateChart::DrawXLabels")
    int             nCurrentIndex = aRangeStart;
    int             nLastIndex    = aRangeStart + aRangeSize;
    double          nLastPos      = XOFFSET;
    int             lastY         = 9999;
    int             lastM         = 9999;
    double          moveLeft      = 0;
    double          yearWidth     = aDC->GetStringWidth ("X") * 2;
    double          monthWidth    = aDC->GetStringWidth ("X") * 1.5;
    double          x1, y, y1, y2;
    MHDate          date;
    MHString        s;
    bool            setNewLastYear = true;

    aDC->SetPenColor (BLACK);
    DrawLine (aXMin, aYMin[aCurrentY], aXMin, aYMax[aCurrentY]);
    DrawLine (aXMax, aYMin[aCurrentY], aXMax, aYMax[aCurrentY]);

    if (!aSettings.aXLabels)
        return;

    y  = TranslateY (aYMin[aCurrentY]) + aCharHeight + SMALL_TICK_LINE_SIZE;
    y1 = TranslateY (aYMin[aCurrentY]) + SMALL_TICK_LINE_SIZE;
    if (aSettings.aVLines)
        y2 = TranslateY (aYMax[aCurrentY]);
    else
        y2 = TranslateY (aYMin[aCurrentY]);

    while (nCurrentIndex < nLastIndex && nCurrentIndex < aDates->Size()) {
        x1 = TranslateX (nCurrentIndex);
        date.Set (((MHString*) (*aDates)[nCurrentIndex])->Get());

        //  Create date string
        s = "";
        if ((aSettings.aTimeType == MHDate::WEEKDAY || aSettings.aTimeType == MHDate::ALL) && lastM != date.GetMonth ()) {
            if (lastM > date.GetMonth () && nLastPos <= (x1 - yearWidth)) {
                date.Get (MHDate::YEAR, s);
                moveLeft = yearWidth;
            }
            else if (lastM < date.GetMonth () && nLastPos <= (x1 - monthWidth)) {
                date.Get (MHDate::ABBR_OF_MONTH, s);
                moveLeft = monthWidth;
            }
            setNewLastYear = true;
        }
        else if ((aSettings.aTimeType == MHDate::FRIDAY || aSettings.aTimeType == MHDate::SUNDAY) && lastY != date.GetLongYear ()) {
            if (nLastPos < (x1 - yearWidth)) {
                date.Get (MHDate::YEAR, s);
                moveLeft = yearWidth;
            }
            setNewLastYear = true;
        }

        //  Make sure prev year is printed at least once
        if (setNewLastYear == true) {
            lastM = date.GetMonth ();
            lastY = date.GetLongYear ();
            setNewLastYear = false;
        }
        else if (lastM == 9999) {
            //  Check when first year is going to be drawn.
            //  If to little days left on year skip that year.
            lastY     = date.GetLongYear ();
            lastM     = date.GetMonth ();
            int days  = (aDC->GetStringWidth ("X") * 5) / aSettings.aXWidth;
            int count = 0;

            for (int f = nCurrentIndex; f < aDates->Size(); f++) {
                date.Set (((MHString*) (*aDates)[f])->Get());
                if (date.GetLongYear () != lastY || count > days) break;
                count++;
            }
            if (count < days && aDates->Size() > count) s = "";
            setNewLastYear = false;
        }

        if (s.Size() > 0) {
            aDC->DrawLine ((int)(x1), (int)(y1), (int)(x1), (int)(y2));
            aDC->DrawText (s(), int(x1 - moveLeft), int(y));
            nLastPos = x1 + moveLeft;
        }
        nCurrentIndex++;
    }
}



/**
*  Go through all dataseries within the range amd get max/min dates and values.
*  @return int - Size of range
*/
int MHDateChart::GetMinMax () {
    DBG_TIME ("MHDateChart::GetMinMax")

    int             nLastIndex  = aRangeStart + aRangeSize;
    int             nCountMax   = 0;
    bool            yinit[2];
    MHPoint*        p;
    MHPoint         p2;

    aYMax[LEFT]  = MAX_VALUE;
    aYMin[LEFT]  = MAX_VALUE;
    aYMax[RIGHT] = MAX_VALUE;
    aYMin[RIGHT] = MAX_VALUE;
    yinit[LEFT]  = false;
    yinit[RIGHT] = false;
    aValid[LEFT] = 0;
    aValid[RIGHT]= 0;
    aXMax        = 0;
    aXMin        = 0;

    //  Iterate through the dateserie
    if (!aDates || aDates->Size() == 0) return 0;

    for (int f = 0; f < MAX_SERIES; f++) {
        int nCurrentIndex = aRangeStart;
        int nCountRange   = 0;
        int nPointIndex   = 0;
        int foundIndex    = 1;

        aCurrentY = aPositions[f];

        #ifdef BUGG
            if (aPoints[f])
                DBG_LOG3 ("%d - %d - %d", nCurrentIndex, nLastIndex, aPoints[f]->Size())
        #endif

        while (aPoints[f] && nCurrentIndex < nLastIndex && nCurrentIndex < aDates->Size() && nCurrentIndex >= 0) {
            p2.SetDate ((MHString*) (*aDates)[nCurrentIndex]);

            if (nPointIndex == 0)
                p = (MHPoint*) aPoints[f]->Find (&p2, nPointIndex, 0);
            else {
                p = (MHPoint*) aPoints[f]->Search (&p2, nPointIndex, 0, &foundIndex);
                if (p) nPointIndex = foundIndex;
            }

            if (p && p->GetClose() < MAX_VALUE && p->GetVolume() < MAX_VALUE) {
                if (yinit[aCurrentY] == false) {
                    //  This is the first data in the first set so it must be max and min
                    aYMax[aCurrentY]  = p->GetHigh();
                    aYMin[aCurrentY]  = p->GetLow();
                    yinit[aCurrentY]  = true;
                    aValid[aCurrentY] = 1;
                }
                else {
                    if (p->GetHigh() > aYMax[aCurrentY]) aYMax[aCurrentY] = p->GetHigh();
                    if (p->GetLow() < aYMin[aCurrentY]) aYMin[aCurrentY] = p->GetLow();
                }
            }
            nCountRange++;
            if (nCountRange > aRangeSize)
                break;

            if (nCountRange > nCountMax)
                nCountMax = nCountRange;
            nCurrentIndex++;
        }
    }
    aXMax = nCountMax + aRangeStart;
    aXMin = aRangeStart;

    return nCountMax;
}



/**
*  Query data point
*  @param int      - X pos
*  @param int      - Y Pos
*  @param MHPoint* - Result point, MHPoint::GetOpen contains current value below mouse
*  @return bool    - true if found
*/
bool MHDateChart::QueryLeft (int x, int y, MHPoint* point) {
    if (aSettings.aChartLeft) {
        aCurrentY   = LEFT;
        double x2   = TranslateX2World(x + 1);
        double y2   = TranslateY2World(y);
        double XMin = TranslateX(aXMin);
        double XMax = TranslateX(aXMax);
        double YMin = TranslateY(aYMin[aCurrentY]);
        double YMax = TranslateY(aYMax[aCurrentY]);

        //  Must be inside chart
        if (aDates && x2 >= 0 && x2 < aDates->Size() && y >= YMax && y <= YMin && x >= XMin && x <= XMax) {
            MHString* s = (MHString*) aDates->At (int(x2));
            point->SetDate (s->Get());
            point->SetHigh (aYMax[aCurrentY]);
            point->SetLow (aYMin[aCurrentY]);
            point->SetClose (y2);
            return true;
        }
    }
    return false;
}



/**
*  Query data point
*  @param int      - X pos
*  @param int      - Y Pos
*  @param MHPoint* - Result point, MHPoint::GetOpen contains current value below mouse
*  @return bool    - true if found
*/
bool MHDateChart::QueryRight (int x, int y, MHPoint* point) {
    if (aSettings.aChartRight) {
        aCurrentY   = RIGHT;
        double x2   = TranslateX2World(x + 1);
        double y2   = TranslateY2World(y);
        double XMin = TranslateX(aXMin);
        double XMax = TranslateX(aXMax);
        double YMin = TranslateY(aYMin[aCurrentY]);
        double YMax = TranslateY(aYMax[aCurrentY]);

        //  Must be inside chart
        if (aDates && x2 >= 0 && x2 < aDates->Size() && y >= YMax && y <= YMin && x >= XMin && x <= XMax) {
            MHString* s = (MHString*) aDates->At (int(x2));
            point->SetDate (s->Get());
            point->SetHigh (aYMax[aCurrentY]);
            point->SetLow (aYMin[aCurrentY]);
            point->SetClose (y2);
            return true;
        }
    }
    return false;
}



/**
*  Set list with closed trading days
*  @param MHList* - List with dates
*/
void MHDateChart::SetBlockList (MHVector* ve) {
    if (aClosed) aClosed->Erase ();
    delete aClosed;
    aClosed = ve;
}



/**
*  Set new min, max dates.
*  @param const char* - Start date
*  @param const char* - Stop date
*/
void MHDateChart::SetDateRange(const char* pStartDate, const char* pStopDate) {
    aXMinDate = pStartDate;
    aXMaxDate = pStopDate;
}
