/**
*  Copyright Mikael Hgdahl - triyana@users.sourceforge.net
*
*  This source is distributed under the terms of the Q Public License version 1.0,
*  created by Trolltech (www.trolltech.com).
*/

#include <math.h>
#include <MHPoint.h>
#include <MHDate.h>



/**
*  Copy constructor
*  @param const MHPoint& - Point to copy from
*/
MHPoint::MHPoint (const MHPoint& x) :
X(x.X),
H(x.H),
L(x.L),
C(x.C),
O(x.O),
V(x.V) {
    strncpy (Date, x.Date, 8);
    Date[8] = '\0';
}



/**
*  Create new data point
*  @param const char* - X value in the form of a date
*  @param double      - Highest value in this data point
*  @param double      - Lowest value in this data point
*  @param double      - Last value in this data point
*  @param double      - Open value in this data point
*  @param double      - Volume value in this data point
*/
MHPoint::MHPoint (const char* d, const double h, const double l, const double c, const double o, const double v) :
X(-1),
H(h),
L(l),
C(c),
O(o),
V(v) {
    strncpy (Date, d, 8);
    Date[8] = '\0';
}



/**
*  Make a vector with the ATR model (Average True Range)
*  @param MHVector* -  Vector with MHPoints
*  @param MHVector* -  The vector with converted data
*  @param int          -  Number of points in range
*/
void MHPoint::ATR (MHVector* in, MHVector* out, int nMvg) {
    int             nSize    = in->Size();
    double          nAtr     = 0;
    double          nTot     = 0;
    double          nPrevC   = 0;
    double          nPrevAtr = 0;
    MHPoint*        p        = 0;

    if (nMvg < 2 || nMvg > in->Size()) return;
    nMvg--;

    //  Iterate through all data points and calc ATR data
    for (int f = 0; f < nSize; f++) {
        p = (MHPoint*)(*in)[f];

        if (f == 0) {
            nAtr = p->H - p->L;
            nTot += nAtr;
        }
        else {
            double t1 = p->H - p->L;
            double t2 = fabs(p->H - nPrevC);
            double t3 = fabs(p->L - nPrevC);

            if (t1 > t2 && t1 > t3) nAtr = t1;
            else if (t2 > t1 && t2 > t3) nAtr = t2;
            else nAtr = t3;

            nTot += nAtr;

            if (f == nMvg) {
                nPrevAtr = nTot / (nMvg + 1);
                out->Push (new MHPoint (p->Date, nPrevAtr));
            }
            else if (f > nMvg) {
                nPrevAtr = ((nPrevAtr * nMvg) + nAtr) / (nMvg + 1);
                out->Push (new MHPoint (p->Date, nPrevAtr));
            }
        }
        nPrevC = p->C;
    }
}



/**
*  Make a vector with the Bollinger model
*  @param MHVector* -  Vector with MHPoints
*  @param MHVector* -  The vector with converted data
*  @param MHVector* -  The vector with converted data + 2 * std. dev.
*  @param MHVector* -  The vector with converted data - 2 * std. dev.
*  @param int          -  Number of points to use in moving average
*/
void MHPoint::Bollinger (MHVector* in, MHVector* mvg, MHVector* upper, MHVector* lower, int nMvg) {
    MHVector stddev;
    MHPoint* ps = 0;
    MHPoint* pm = 0;

    MHPoint::MovingAverage (in, mvg, nMvg);
    MHPoint::StdDev (in, &stddev, nMvg);

    //  Create two new dataseries with mvg +/- standard deviations * 2
    for (int i = 0; i < stddev.Size(); i++) {
        ps = (MHPoint*)stddev[i];
        pm = (MHPoint*)mvg->At (i);

        if (pm) {
            upper->Push (new MHPoint (ps->Date, pm->C + (ps->C) * 2));
            lower->Push (new MHPoint (ps->Date, pm->C - (ps->C) * 2));
        }
    }
    stddev.Erase();
}



/**
*  Copy points to new vector, create new points
*  @param MHVector*    - In serie
*  @param MHVector*    - Out serie
*  @param const char*  - Start date, if 0 then all data is copied
*  @param const char*  - Stop date, if 0 then all data is copied
*/
void MHPoint::Copy (MHVector* in, MHVector* out, const char* pStartDate, const char* pStopDate) {
    MHPoint* p;
    for (int f = 0; f < in->Size(); f++) {
        p = (MHPoint*)(*in)[f];
        if (pStartDate == 0 || pStopDate == 0)
            out->Push (new MHPoint (*p));
        else if (*p >= pStartDate && *p <= pStopDate)
            out->Push (new MHPoint (*p));
    }
}



/**
*  Copy pointers to points to new vector
*  @param MHVector*    - In serie
*  @param MHVector*    - Out serie
*/
void MHPoint::CopyPointers (MHVector* in, MHVector* out) {
    for (int f = 0; f < in->Size(); f++)
        out->Push ((MHPoint*)(*in)[f]);
}



/**
*  Fix slave list so it's range is the same as master
*  @param MHVector* - In serie
*  @param MHVector* - Out serie
*/
void MHPoint::Cut (MHVector* in, MHVector* out) {
    MHPoint*    p1     = (MHPoint*)in->First();
    MHPoint*    p2     = (MHPoint*)in->Last();
    MHPoint*    p;
    MHVector    v;

    if (!p1 || !p2) return;

    for (int f = 0; f < out->Size(); f++) {
        p = (MHPoint*) (*out)[f];

        if (*p >= *p1 && *p <= *p2) {
            //  First date must be the same
            if (v.Size() == 0 && *p != *p1)
                v.Push (new MHPoint(*p1));
            v.Push (new MHPoint(*p));
        }
    }
    //  Last date must be the same
    p = (MHPoint*) v.Last();
    if (p && *p != *p2)
        v.Push (new MHPoint(*p2));

    out->Erase();
    MHPoint::CopyPointers (&v, out);
    v.Empty();
}



/**
*  Convert day price data to weekly data
*  @param MHVector* - In serie
*  @param MHVector* - Out serie
*  @param int       -  One of MHDate::MONDAY - MHDate::SUNDAY
*/
void MHPoint::Day2Week (MHVector* in, MHVector* out, int nWeekDay) {
    bool        Init = false;
    MHString    StopStr, StartStr;
    MHPoint*    p;
    MHPoint     p2;
    MHDate      StopDate;

    for (int f = 0; f < in->Size(); f++) {
        p = (MHPoint*) (*in)[f];

        if (Init == false) {
            //  Initiate first date.
            //  Move stopdate to right day in week
            MHDate tmp;

            tmp.Set (p->Date);
            tmp.GotoWeekday (nWeekDay);
            tmp.Get (MHDate::LONG_DATE, StopStr);
            StopDate = tmp;

            //  Then move start date back one week
            tmp.AddDays (-7);
            tmp.Get (MHDate::LONG_DATE, StartStr);

            p2 = *p;
            Init = true;
        }

        if (*p > StopStr()) {
            //  Current point is out of range so save old point to out vector
            p2.SetDate (StopStr());
            out->Push (new MHPoint(p2));

            //  Set current point as last point
            p2 = *p;
        }
        else if (*p > StartStr() && *p <= StopStr()) {
            //  Save last point for every time serie range
            if (p2 == "")
                p2 = *p;
            else {
                if (p->H > p2.H) p2.H = p->H;
                if (p->L < p2.L) p2.L = p->L;
                p2.V += p->V;
                p2.C = p->C;
            }
        }

        //  Set new range if point is outside current range
        while (*p > StopStr()) {
            StartStr = StopStr;
            StopDate.AddDays (7);
            StopDate.Get (MHDate::LONG_DATE, StopStr);
        }
    }
    if (StopStr != MHString("")) {
        p2.SetDate (StopStr());
        out->Push (new MHPoint(p2));
    }
}



/**
*  Make a vector with the Momentum model
*  @param MHVector* -  Vector with MHPoints
*  @param MHVector* -  The vector with converted data
*  @param MHVector* -  The vector with zero line
*  @param int          -  Number of days back in time
*/
void MHPoint::Momentum (MHVector* in, MHVector* out, MHVector* zero, int nMvg) {
    int      nSize  = in->Size();
    int      nStart = nMvg - 1;
    MHPoint* p;
    MHPoint* p2;

    if (nMvg < 2 && nMvg > nSize)
        return;

    //  Iterate through all data points and calc RSI data
    for (int f = 1; f < nSize; f++) {
        if (f >= nStart) {
            p  = (MHPoint*) (*in)[f];
            p2 = (MHPoint*) (*in)[f - nStart];
            out->Push (new MHPoint (p->Date, p->C - p2->C));
            zero->Push (new MHPoint (p->Date, 0));
        }
    }
}



/**
*  Make a vector with moving average of price data
*  @param MHVector* - In serie
*  @param MHVector* - Out serie
*  @param int       -  Number of points in moving average
*/
void MHPoint::MovingAverage (MHVector* in, MHVector* out, int nMvg) {
    int             nCount = 0;
    double          nSum   = 0;
    double*         v = new double [in->Size()];
    MHPoint*        p;

    if (nMvg > 2 && nMvg < in->Size()) {
        //  Iterate through all data points
        for (int f = 0; f < in->Size(); f++) {
            p = (MHPoint*) (*in)[f];
            nCount++;

            if (nCount < nMvg) {
                //  Add data until the first moving average price can be calculated
                v[nCount - 1] = p->C;
                nSum += p->C;
            }
            else if (nCount == nMvg) {
                //  This is the first point
                v[nCount - 1] = p->C;
                nSum += p->C;
                out->Push (new MHPoint (p->Date, nSum / nMvg));
            }
            else {
                //  And here comes the rest
                v[nCount - 1] = p->C;

                //  Remove oldest data in range and add current to sum
                nSum -= v[nCount - (nMvg + 1)];
                nSum += p->C;

                //  save new point, (point 2 and forward)
                out->Push (new MHPoint (p->Date, nSum / nMvg));
            }
        }
    }
    delete []v;
}



/**
*  Create a list with MHPoints that are normalised with
*  the first data point as the reference point
*  @param MHVector*  -  Vector with MHPoints
*  @param MHVector*  -  The vector with converted data
*  @param const double& -  Start number
*/
void MHPoint::Normalise (MHVector* in, MHVector* out, double nOrigo) {
    bool     bFirst  = true;
    double   nFactor = 0;
    MHPoint* p;

    for (int f = 0; f < in->Size(); f++) {
        p = (MHPoint*) (*in)[f];
        if (p->C < 0)
            continue;

        if (bFirst == true) {
            nFactor = p->C / nOrigo;
            bFirst = false;
        }
        out->Push(new MHPoint (
            p->Date,
            p->H / nFactor,
            p->L / nFactor,
            p->C / nFactor,
            p->O / nFactor,
            p->V / nFactor));
    }
}



/**
*  Print to outstream all data points for a vector
*  @param MHVector* - In serie
*/
void MHPoint::Print (MHVector* in) {
    for (int f = 0; f < in->Size(); f++)
        ;//cout << *((MHPoint*)(*in)[f]) << endl;
}



/**
*  Make a vector with the Relative Strength Index
*  @param MHVector* -  Vector with MHPoints
*  @param MHVector* -  The vector with converted data
*  @param int          -  Number of points to use
*/
void MHPoint::RSI (MHVector* in, MHVector* out, int nMvg) {
    int         nStart = nMvg - 1;
    double      nPos   = 0;
    double      nNeg   = 0;
    double      nDiff;
    MHPoint*    p;
    MHPoint*    pp;
    MHPoint*    po;
    MHPoint*    po2;

    if (nMvg < 2 || nMvg > in->Size()) return;

    //  Iterate through all data points and calc RSI data
    for (int f = 1; f < in->Size(); f++) {
        p   = (MHPoint*) (*in)[f];
        pp  = (MHPoint*) (*in)[f - 1];
        po  = (MHPoint*) in->At ((f + 1) - nStart);
        po2 = (MHPoint*) in->At (f - nStart);

        nDiff = p->H - pp->H;
        if (nDiff > 0)
            nPos += nDiff;
        else
            nNeg -= nDiff;

        if (f >= nStart) {
            out->Push (new MHPoint (p->Date, (nPos / (nPos + nNeg)) * 100));

            //  Remove the oldest value in the timeserie
            if (po && po2)
                nDiff = po->H - po2->H;
            if (nDiff > 0)
                nPos -= nDiff;
            else
                nNeg += nDiff;
        }
    }
}



/**
*  Make a vector with the Relative Strength Index
*  @param MHVector* -  Vector with MHPoints
*  @param MHVector* -  The vector with converted data
*  @param int          -  Number of points to use
*/
void MHPoint::RSI2 (MHVector* in, MHVector* out, int nMvg) {
    double      nAvgGain = 0;
    double      nAvgLoss = 0;
    double      nDiff;
    MHPoint*    p;
    MHPoint*    pp;

    if (nMvg < 2 || nMvg > in->Size()) return;

    //  Iterate through all data points and calc RSI data
    for (int f = 1; f < in->Size(); f++) {
        p     = (MHPoint*) (*in)[f];
        pp    = (MHPoint*) (*in)[f - 1];
        nDiff = p->C - pp->C;

        if (f <= nMvg) {
            if (nDiff > 0)
                nAvgGain += nDiff;
            else
                nAvgLoss += fabs(nDiff);
        }

        if (f == nMvg) {
            nAvgGain = nAvgGain / nMvg;
            nAvgLoss = nAvgLoss / nMvg;
            out->Push (new MHPoint (p->Date, 100 - (100 / (1 + (nAvgGain / nAvgLoss)))));
        }
        else if (f > nMvg) {
            nAvgGain = ((nAvgGain * (nMvg - 1)) + ((nDiff > 0) ? fabs(nDiff) : 0)) / nMvg;
            nAvgLoss = ((nAvgLoss * (nMvg - 1)) + ((nDiff < 0) ? fabs(nDiff) : 0)) / nMvg;
            out->Push (new MHPoint (p->Date, 100 - (100 / (1 + (nAvgGain / nAvgLoss)))));
        }
    }
}



/**
*  Save points to file
*  @param MHVector* - In serie
*  @param const char*  - The filename
*  @return bool        - true if ok
*/
bool MHPoint::Save (const char* pFilename, MHVector* in) {
    FILE*       file  = fopen (pFilename, "wb");
    char        buffer[201];
    MHPoint*    p;

    if (file) {
        for (int f = 0; f < in->Size(); f++) {
            p = (MHPoint*)(*in)[f];
            SNPRINTF (buffer, 200, "%s\t%f\t%f\t%f\t%f\t%f\n", p->Date, p->H, p->L, p->C, p->O, p->V);
            fwrite (buffer, sizeof(char), strlen(buffer), file);
        }
        fclose (file);
        return true;
    }
    return false;
}



/**
*  Make a vector with data converted to standar deviation
*  @param MHVector* - In serie
*  @param MHVector* - Out serie
*  @param int          -  Number of points for average
*/
void MHPoint::StdDev (MHVector* in, MHVector* out, int nDays) {
    int       nCount = 0;
    double    nMean  = 0;
    double    nDev   = 0;
    double    nDev2  = 0;
    double    nSum   = 0;
    MHPoint*  p;

    if (nDays < 2 || nDays > in->Size()) return;

    for (int f = 0; f < in->Size(); f++) {
        p = (MHPoint*)(*in)[f];
        nCount++;
        nSum += p->C;

        if (nCount >= nDays) {
            nMean = nSum / nDays;

            //  Create nDays points
            nDev2 = 0;
            for (int j = nCount - nDays; j < nCount; j++) {
                p = (MHPoint*) (*in)[j];
                nDev   = p->C - nMean;
                nDev  *= nDev;
                nDev2 += nDev;
            }
            nDev2 /= nDays;
            nDev2 = sqrt (nDev2);

            p = (MHPoint*) (*in)[nCount - 1];
            out->Push (new MHPoint (p->Date, nDev2));

            p = (MHPoint*) (*in)[nCount - nDays];
            nSum -= p->C;
        }
    }
}



/**
*  Make a vector with the stochastics formula
*  @param MHVector* - In serie
*  @param MHVector* - Out serie
*  @param int          -  Number of points in moving average
*/
void MHPoint::Stochastics (MHVector* in, MHVector* out, int nMvg) {
    int      nCount = 0;
    double   nH, nL, nK, tmp;
    MHPoint* p;
    MHPoint* p2;

    for (int f = 0; f < in->Size(); f++) {
        p = (MHPoint*)(*in)[f];
        nCount++;

        if (nCount >= nMvg) {
            nH = p->H;
            nL = p->L;

            //  Get max/min prices for current range
            for (int j = (nCount - nMvg); j < (nCount - 1); j++) {
                p2 = (MHPoint*) (*in)[j];
                if (p2->H > nH) nH = p2->H;
                if (p2->L < nL) nL = p2->L;
            }

            tmp = nH - nL;
            nK = (tmp < 0.00005) ? 0.0 : ((p->C - nL) / tmp) * 100;
            //  Hmm, some ugly floating point math error
            if (fabs (nK) < 0.0001) nK = 0+0.000001;
            if (fabs(nK-100.0) < 0.0001) nK = 100-0.000001;
            out->Push (new MHPoint (p->Date, nK));
        }
    }
}



/**
*  Validate point and try to fix missing data
*/
void MHPoint::Validate () {
    MHDate now;
    if (strlen(Date) != 8) SetDate (now.Get (MHDate::LONG_DATE));
    if (C > H) H = C;
    if (C < L) L = C;
    if (O < 0.0000001 && O > -0.0000001) O = -1;
    if (V < 0.0000001 && V > -0.0000001) V = -1;
    if (C < 0.0000001 && C > -0.0000001) {
        H = L =C = -1;
    }
}



/**
*  Make a vector with only volume data
*  @param MHVector* -  Vector with MHPoints
*  @param MHVector* -  The vector with converted data
*/
void MHPoint::Volume (MHVector* in, MHVector* out) {
    MHPoint* p;

    for (int f = 0; f < in->Size(); f++) {
        p = (MHPoint*) (*in)[f];
        if (p->V >= 0.0000001)
            out->Push (new MHPoint (p->Date, p->V));
    }
}
