/**
*  Copyright Mikael Hgdahl - triyana@users.sourceforge.com
*
*  This source is distributed under the terms of the Q Public License version 1.0,
*  created by Trolltech (www.trolltech.com).
*/

#include "Fl_DC.h"
#include <MHChart.h>
#include <MHDebug.h>



/**
*  Create a chart canvas.
*  @param int   - x pos
*  @param int   - y pos
*  @param int   - width
*  @param int   - height
*  @param char* - label
*/
Fl_DC::Fl_DC (int x, int y, int w, int h, char *label) :
Fl_Group (x, y, w, h, label) {
    aCanvas    = new Fl_Box (x, y, w, h - SCROLL_SIZE);
    aScrollbar = new Fl_Scrollbar (x, y + h - SCROLL_HEIGHT, w, SCROLL_HEIGHT);

    aScrollbar->type (FL_HORIZONTAL);
    aScrollbar->value (1, 1, 1, 1);
    aCanvas->box (FL_DOWN_BOX);
    aCanvas->color (FL_WHITE);

    add (aCanvas);
    add (aScrollbar);

    resizable (0);
}



/**
*  Destroy widget and also clear all data
*/
Fl_DC::~Fl_DC () {
    Clear();
}



/**
*  All events goes here
*  @param int  - Event
*  @return int - 1 for handled event 0 for not
*/
int Fl_DC::handle (int event) {
    return Fl_Group::handle (event);
}


/**
*  Resize only canvas but not the charts
*  @param int   - x pos
*  @param int   - y pos
*  @param int   - width
*  @param int   - height
*/
void Fl_DC::resize (int x, int y, int w, int h) {
    Fl_Group::resize(x, y, w, h);
    aCanvas->resize(x, y, w, h - SCROLL_SIZE);
    aScrollbar->resize(x, y + h - SCROLL_HEIGHT, w, SCROLL_HEIGHT);

    if (w < 0 || h < 0) return;
    if (w == aW && h == aH) return;
    aW = w;
    aH = h;

    Recalc (true, this->w(), this->h());
}



/**
*  Recalc chart data
*  @param bool - set to true if data serie has been added/changed
*  @param int  - Width of canvas, default -1, no resize is done
*  @param int  - Height of canvas, default -1, no resize is done
*/
void Fl_DC::Recalc (bool bInit, int w, int h) {
    fl_font (FL_COURIER, GetFontSize());
    MHDC::Recalc (bInit, w - SPACE, h - (SCROLL_SIZE + SPACE));
}



/**
*  Set color of background canvas
*  @param int - Color constant
*/
void Fl_DC::SetBrushColor (int Color) {
    switch (Color) {
        case MHChart::BLACK:
            fl_color (FL_BLACK);
            break;

        case MHChart::BLUE:
            fl_color (FL_BLUE);
            break;

        case MHChart::RED:
            fl_color (FL_RED);
            break;

        case MHChart::GREEN:
            fl_color (fl_darker(FL_GREEN));
            break;

        case MHChart::MAGENTA:
            fl_color (FL_MAGENTA);
            break;

        case MHChart::GRAY:
            fl_color (FL_GRAY);
            break;

        case MHChart::CYAN:
            fl_color (FL_CYAN);
            break;

        case MHChart::WHITE:
            fl_color (FL_WHITE);
            break;

        case MHChart::YELLOW:
            fl_color (fl_darker(FL_YELLOW));
            break;

        default:
            fl_color (Color);
    }
}



/**
*  Set pencil color to use for all drawings, both lines and text
*  @param int - Color constant
*/
void Fl_DC::SetPenColor (int Color) {
    switch (Color) {
        case MHChart::BLACK:
            fl_color (FL_BLACK);
            break;

        case MHChart::RED:
            fl_color (FL_RED);
            break;

        case MHChart::GREEN:
            fl_color (fl_darker(FL_GREEN));
            break;

        case MHChart::BLUE:
            fl_color (FL_BLUE);
            break;

        case MHChart::MAGENTA:
            fl_color (FL_MAGENTA);
            break;

        case MHChart::CYAN:
            fl_color (FL_CYAN);
            break;

        case MHChart::GRAY:
            fl_color (FL_GRAY);
            break;

        case MHChart::WHITE:
            fl_color (FL_WHITE);
            break;

        case MHChart::YELLOW:
            fl_color (fl_darker(FL_YELLOW));
            break;

        default:
            fl_color (Color);
    }
}



/**
*  Set pos of scrollbar, both in chart and the scrollbar
*  @param int - New pos to set
*/
void Fl_DC::SetScrollPos (int Pos) {
    if (GetSettings(1) && GetSettings(1)->aShowAll) {
        SetStartInDataSet (0);
        ((Fl_Valuator*)aScrollbar)->value (0);
    }
    else {
        SetStartInDataSet (Pos);
        ((Fl_Valuator*)aScrollbar)->value (Pos);
    }
}



/**
*  Set scrollbar range. The maximum size for data serie
*/
void Fl_DC::SetScrollRange () {

    if (GetSettings(1) && GetSettings(1)->aShowAll) {
        aScrollbar->minimum (0);
        aScrollbar->maximum (0);
        aScrollbar->slider_size (0.08);
        aScrollbar->deactivate();
        SetScrollPos (0);
    }
    else {
        int size = GetLargestDataSet();
        aScrollbar->minimum (0);
        aScrollbar->maximum (size);
        aScrollbar->slider_size (0.08);
        aScrollbar->activate();
        SetScrollPos (size);
    }
}
