/**
*  Copyright Mikael Hgdahl - triyana@users.sourceforge.net
*
*  This source is distributed under the terms of the Q Public License version 1.0,
*  created by Trolltech (www.trolltech.com).
*/

#include <FL/Fl.h>
#include <FL/fl_draw.h>
#include <fl/Fl_Layout.h>
#include <stdlib.h>
#include <stdio.h>



/**
*  Create new border layout
*  @param LAYOUT - BORDER, HORIZONTAL, HORIZONTAL_SPLIT, VERTICAL, VERTICAL_SPLIT
*  @param int    - X pos
*  @param int    - Y pos
*  @param int    - Width
*  @param int    - Height
*/
Fl_Layout::Fl_Layout (LAYOUT layout, int x, int y, int w, int h) :
Fl_Group (x, y, w, h, 0) {
    aLayout          = layout;
    aWidgets[NORTH]  = 0;
    aWidgets[SOUTH]  = 0;
    aWidgets[WEST]   = 0;
    aWidgets[EAST]   = 0;
    aWidgets[CENTER] = 0;
    aWidget1         = 0;
    aWidget2         = 0;
    aSize1           = 0;
    aSize2           = 0;
    aDrag            = 0;
    aSpace           = 0;
    aNorth           = 0;
    aSouth           = 0;
    aWest            = 0;
    aEast            = 0;

    resizable (0);
}



/**
*  Add child widget
*/
void Fl_Layout::add (Fl_Widget* widget) {
    if (aLayout == HORIZONTAL || aLayout == VERTICAL)
        Fl_Group::add (widget);
}



/**
*  Add child widget
*/
void Fl_Layout::add (POSITION position, Fl_Widget* widget) {
    if (aLayout == BORDER) {
        switch (position) {
            case NORTH:
                if (aWidgets[NORTH]) remove (aWidgets[NORTH]);
                delete aWidgets[NORTH];
                Fl_Group::add (widget);
                aWidgets[NORTH] = widget;
                break;

            case SOUTH:
                if (aWidgets[SOUTH]) remove (aWidgets[SOUTH]);
                delete aWidgets[SOUTH];
                Fl_Group::add (widget);
                aWidgets[SOUTH] = widget;
                break;

            case WEST:
                if (aWidgets[WEST]) remove (aWidgets[WEST]);
                delete aWidgets[WEST];
                Fl_Group::add (widget);
                aWidgets[WEST] = widget;
                break;

            case EAST:
                if (aWidgets[EAST]) remove (aWidgets[EAST]);
                delete aWidgets[EAST];
                Fl_Group::add (widget);
                aWidgets[EAST] = widget;
                break;

            case CENTER:
                if (aWidgets[CENTER]) remove (aWidgets[CENTER]);
                delete aWidgets[CENTER];
                Fl_Group::add (widget);
                aWidgets[CENTER] = widget;
                break;

            default:
                exit(0);
        }
    }
    else if (aLayout == HORIZONTAL_SPLIT) {
        switch (position) {
            case WEST:
                if (aWidget1) remove (aWidget1);
                delete aWidget1;
                Fl_Group::add (widget);
                aWidget1 = widget;
                aSize1   = widget->w();
                break;

            case EAST:
                if (aWidget2) remove (aWidget2);
                delete aWidget2;
                Fl_Group::add (widget);
                aWidget2 = widget;
                aSize2   = widget->w();
                break;

            default:
                exit(0);
        }
    }
    else if (aLayout == VERTICAL_SPLIT) {
        switch (position) {
            case NORTH:
                if (aWidget1) remove (aWidget1);
                delete aWidget1;
                Fl_Group::add (widget);
                aWidget1 = widget;
                aSize1   = widget->h();
                break;

            case SOUTH:
                if (aWidget2) remove (aWidget2);
                delete aWidget2;
                Fl_Group::add (widget);
                aWidget2 = widget;
                aSize2   = widget->h();
                break;

            default:
                exit(0);
        }
    }
    else
        exit(0);
}



/**
*  All events goes here, but only for HORIZONTAL_SPLIT and VERTICA_SPLIT
*  @param int  - Event
*  @return int - 1 for handled event 0 for not
*/
int Fl_Layout::handle (int event) {
    if (!aWidget1 || !aWidget2) return Fl_Group::handle (event);
    if (aSize1 == 0 || aSize2 == 0) return Fl_Group::handle (event);

    switch (event) {
        case FL_DRAG:
            if (aDrag == HORIZONTAL_SPLIT) {
                if (aWidget1->w() > 0) aSize1 = Fl::event_x() - x();
                if (aSize1 <= 0) aSize1 = 1;
                if (aSize1 > (w() - 0)) aSize1 = w() - 0;
                aWidget1->resize (0, 0, 0, 0);
                resize (x(), y(), w(), h());
                Fl::redraw ();
            }
            else if (aDrag == VERTICAL_SPLIT) {
                if (aWidget1->h() > 0) aSize1 = Fl::event_y() - y();
                if (aSize1 <= 0) aSize1 = 1;
                if (aSize1 > (h() - 0)) aSize1 = h() - 0;
                aWidget1->resize (0, 0, 0, 0);
                resize (x(), y(), w(), h());
                Fl::redraw ();
            }
            return 1;

        case FL_LEAVE:
            fl_cursor (FL_CURSOR_DEFAULT);
            aDrag = 0;
            return 0;

        case FL_MOVE:
            if (aLayout == HORIZONTAL_SPLIT) {
                if (Fl::event_x() > (aWidget1->x() + aWidget1->w() - 4) && Fl::event_x() < (aWidget1->x() + aWidget1->w() + 2)) {
                    fl_cursor (FL_CURSOR_WE);
                    aDrag = HORIZONTAL_SPLIT;
                }
                else {
                    fl_cursor (FL_CURSOR_DEFAULT);
                    aDrag = 0;
                }
            }
            else if (aLayout == VERTICAL_SPLIT) {
                if (Fl::event_y() > (aWidget1->y() + aWidget1->h() - 4) && Fl::event_y() < (aWidget1->y() + aWidget1->h() + 2)) {
                    fl_cursor (FL_CURSOR_NS);
                    aDrag = VERTICAL_SPLIT;
                }
                else {
                    fl_cursor (FL_CURSOR_DEFAULT);
                    aDrag = 0;
                }
            }
            Fl_Group::handle (event);
            return 0;

        case FL_PUSH:
            if (!aDrag)
                Fl_Group::handle (event);
            return 1;

        default:
            return Fl_Group::handle (event);
    }
}



/**
*  Resize layout and all children
*  @param int  - X pos
*  @param int  - Y pos
*  @param int  - Width
*  @param int  - Height
*/
void Fl_Layout::resize (int xpos, int ypos, int width, int height) {
    Fl_Group::resize (xpos, ypos, width, height);

    xpos   += aWest;
    ypos   += aNorth;
    width  -= aWest;
    width  -= aEast;
    height -= aNorth;
    height -= aSouth;

    switch (aLayout) {
        case BORDER:
            resizeBorder (xpos, ypos, width, height);
            break;

        case HORIZONTAL:
            resizeHorizontal (xpos, ypos, width, height);
            break;

        case HORIZONTAL_SPLIT:
            resizeHorizontalSplit (xpos, ypos, width, height);
            break;

        case VERTICAL:
            resizeVertical (xpos, ypos, width, height);
            break;

        case VERTICAL_SPLIT:
            resizeVerticalSplit (xpos, ypos, width, height);
            break;
    }
}



/**
*  Resize layout and all children
*  @param int  - X pos
*  @param int  - Y pos
*  @param int  - Width
*  @param int  - Height
*/
void Fl_Layout::resizeBorder (int xpos, int ypos, int width, int height) {
    if (aWidgets[NORTH]) {
        aWidgets[NORTH]->resize (xpos, ypos, width, aWidgets[NORTH]->h());
        ypos += aWidgets[NORTH]->h();
        height -= aWidgets[NORTH]->h();
    }

    if (aWidgets[SOUTH]) {
        aWidgets[SOUTH]->resize (xpos, y() + h() - aWidgets[SOUTH]->h(), width, aWidgets[SOUTH]->h());
        height -= aWidgets[SOUTH]->h();
    }

    if (aWidgets[WEST]) {
        aWidgets[WEST]->resize (xpos, ypos, aWidgets[WEST]->w(), height);
        xpos += aWidgets[WEST]->w();
        width -= aWidgets[WEST]->w();
    }

    if (aWidgets[EAST]) {
        aWidgets[EAST]->resize (x() + w() - aWidgets[EAST]->w(), ypos, aWidgets[EAST]->w(), height);
        width -= aWidgets[EAST]->w();
    }

    if (aWidgets[CENTER]) {
        aWidgets[CENTER]->resize (xpos, ypos + aSpace, width, height - aSpace);
    }
}



/**
*  Resize horizontal layout and all children, set children height to layout height
*  @param int  - X pos
*  @param int  - Y pos
*  @param int  - Width
*  @param int  - Height
*/
void Fl_Layout::resizeHorizontal (int xpos, int ypos, int width, int height) {
    for (int f = 0; f < children(); f++) {
        Fl_Widget* w = child (f);

        if ((xpos + w->w()) > (x() + width))
            w->hide ();
        else
            w->show ();
        w->resize (xpos, ypos, w->w(), height);
        xpos += w->w() + aSpace;
    }
}



/**
*  Resize horizontal split layout and its two child widgets
*  @param int  - X pos
*  @param int  - Y pos
*  @param int  - Width
*  @param int  - Height
*/
void Fl_Layout::resizeHorizontalSplit (int xpos, int ypos, int width, int height) {
    if (aWidget1 && aWidget2) {
        aWidget1->show();
        aWidget2->show();

        if (aWidget1->w() > width)
            aWidget1->resize (xpos, ypos, width, height);
        else if (aWidget1->w() < aSize1 && width >= aSize1)
            aWidget1->resize (xpos, ypos, aSize1, height);
        else if (aWidget1->w() < aSize1 && width < aSize1)
            aWidget1->resize (xpos, ypos, width, height);
        else
            aWidget1->resize (xpos, ypos, aWidget1->w(), height);

        if ((aWidget1->w() + aWidget2->w() + aSpace) > width)
            aWidget2->resize (xpos + aWidget1->w() + aSpace, ypos, width - aWidget1->w() - aSpace, height);
        else
            aWidget2->resize (xpos + aWidget1->w() + aSpace, ypos, width - aWidget1->w() - aSpace, height);

        if (aWidget1->w() <= 0) aWidget1->hide();
        if (aWidget2->w() <= 0) aWidget2->hide();
    }
}



/**
*  Resize vertical layout and all children, set children width to layout width
*  @param int  - X pos
*  @param int  - Y pos
*  @param int  - Width
*  @param int  - Height
*/
void Fl_Layout::resizeVertical (int xpos, int ypos, int width, int height) {
    for (int f = 0; f < children(); f++) {
        Fl_Widget* w = child (f);

        if ((ypos + w->h()) > (y() + height))
            w->hide ();
        else
            w->show ();

        w->resize (xpos, ypos, width, w->h());
        ypos += w->h() + aSpace;
    }
}



/**
*  Resize vertical split layout and its two child widgets
*  @param int  - X pos
*  @param int  - Y pos
*  @param int  - Width
*  @param int  - Height
*/
void Fl_Layout::resizeVerticalSplit (int xpos, int ypos, int width, int height) {
    if (aWidget1 && aWidget2) {
        aWidget1->show();
        aWidget2->show();

        if (aSize2 > 0) {
            if (aWidget1->h() > height)
                aWidget1->resize (xpos, ypos, width, height);
            else if (aWidget1->h() < aSize1 && height >= aSize1)
                aWidget1->resize (xpos, ypos, width, aSize1);
            else if (aWidget1->h() < aSize1 && height < aSize1)
                aWidget1->resize (xpos, ypos, width, height);
            else
                aWidget1->resize (xpos, ypos, width, aWidget1->h());

            if ((aWidget1->h() + aWidget2->h() + aSpace) > height)
                aWidget2->resize (xpos, ypos + aWidget1->h() + aSpace, width, height - aWidget1->h() - aSpace);
            else
                aWidget2->resize (xpos, ypos + aWidget1->h() + aSpace, width, height - aWidget1->h()- aSpace);
        }
        else
            aWidget1->resize (xpos, ypos, width, height);

        if (aWidget1->h() <= 0) aWidget1->hide();
        if (aWidget2->h() <= 0) aWidget2->hide();
    }
}
